import unittest
import cgi

class SERVER_RETURN(Exception): pass
class APACHE_EXCEPTION(Exception): pass
def dummyApache():
    """Create a dummy _apache module to make it so that we can import
    mod_python's apache module."""
    import new
    _apache = new.module('_apache')
    _apache.table = None
    _apache.log_error = None
    _apache.config_tree = None
    _apache.server_root = None
    _apache.mpm_query = lambda x: False
    _apache.exists_config_define = None
    _apache.SERVER_RETURN = SERVER_RETURN
    for exc in ['APR_NOFILE', 'APR_REG', 'APR_DIR', 'APR_CHR', 'APR_BLK',
                'APR_PIPE', 'APR_LNK', 'APR_SOCK', 'APR_UNKFILE',
                'AP_CONN_UNKNOWN', 'AP_CONN_CLOSE', 'AP_CONN_KEEPALIVE']:
        setattr(_apache, exc, APACHE_EXCEPTION)
    _apache.parse_qs = cgi.parse_qs
    _apache.parse_qsl = cgi.parse_qsl
    _apache.stat = None
    import sys
    sys.modules['_apache'] = _apache
    import _apache
    return _apache

dummyApache()
import mpopenid

###### Authorized users

def mkAuthTest(case, expected):
    def test(self):
        actual = mpopenid.parse_authorized_urls(case)
        self.failUnlessEqual(
            actual,
            expected,
            'Case %r, got %r, expected %r' % (case, actual, expected))
    return test

class ParseAuthorizedTest(unittest.TestCase):
    test_empty = mkAuthTest('', [])
    test_empty2 = mkAuthTest(' ', [])
    test_empty3 = mkAuthTest('\n\n', [])
    test_one = mkAuthTest('http://www.jyte.com/', ['http://www.jyte.com/'])
    test_one_trail = mkAuthTest('http://www.jyte.com', ['http://www.jyte.com/'])
    test_one_trail_pfx = mkAuthTest('www.jyte.com', ['http://www.jyte.com/'])
    test_one_pfx = mkAuthTest('www.jyte.com/', ['http://www.jyte.com/'])
    test_two = mkAuthTest('http://one.com/ http://two.com/',
                          ['http://one.com/', 'http://two.com/'])
    test_two_nl = mkAuthTest('http://one.com/\nhttp://two.com/',
                             ['http://one.com/', 'http://two.com/'])
    test_two_leadtrail = mkAuthTest(' http://one.com/\nhttp://two.com/ ',
                                    ['http://one.com/', 'http://two.com/'])
    test_two_pfx = mkAuthTest('one.com/ two.com/',
                              ['http://one.com/', 'http://two.com/'])

    test_file_list = mkAuthTest('''\
http://j3h.livejournal.com/
http://j3h.janrain.com/
http://j3h.us/

http://mylid.net/j3h.

http://josh.utest.magbox.com/

http://josh.myopenid.com/
http://j3h.myopenid.com/
''', [
        'http://j3h.livejournal.com/',
        'http://j3h.janrain.com/',
        'http://j3h.us/',
        'http://mylid.net/j3h.',
        'http://josh.utest.magbox.com/',
        'http://josh.myopenid.com/',
        'http://j3h.myopenid.com/',
      ])

###### Cookied User

def mkRequest(cls, session, options, req):
    self = object.__new__(cls)
    self.session = session
    self.options = options
    self.apache_request = req
    return self

class CookiedUserTest(unittest.TestCase):
    def test_missing(self):
        req = mkRequest(mpopenid.OpenIDAccessRequest, {}, None, None)
        self.failUnless(req.cookied_user is None)

    def test_present(self):
        session = {
            'cookied_user': 'http://j3h.us/',
            }
        req = mkRequest(mpopenid.OpenIDAccessRequest, session, None, None)
        self.failUnlessEqual(req.cookied_user, 'http://j3h.us/')
        self.failUnlessEqual(req.get_cookied_user(), 'http://j3h.us/')

    def test_set(self):
        req = mkRequest(mpopenid.OpenIDAccessRequest, {}, None, None)
        self.failUnless(req.cookied_user is None)
        req.set_cookied_user('http://josh.myopenid.com/')
        self.failUnlessEqual(req.cookied_user, 'http://josh.myopenid.com/')
        req.cookied_user = 'http://j3h.us/'
        self.failUnlessEqual(req.cookied_user, 'http://j3h.us/')
        self.failUnlessEqual(req.get_cookied_user(), 'http://j3h.us/')
        self.failUnlessEqual(req.session.get('cookied_user'), 'http://j3h.us/')

###### Server URL

class Dummy(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

class DummyRequest(Dummy):
    hostname = 'localhost'
    path_info = ''
    uri = ''
    method = 'GET'
    def __init__(self, **kwargs):
        self.headers_out = DummyTable()
        self.headers_in = DummyTable()
        self.connection = Dummy(local_addr=(None, 80))
        super(DummyRequest, self).__init__(**kwargs)
    def is_https(self):
        return getattr(self, 'https', False)
    def set_content_length(self, clen):
        self.content_length = clen
    def write(self, text):
        self.text = text

def mkServerURLTest(case, expected):
    def test(self):
        req = self._request(*case)
        actual = req.server_url
        self.failUnlessEqual(actual, expected)

    return test

class ServerURLTest(unittest.TestCase):
    def _request(self, hostname, local_port, https=None):
        req = DummyRequest(hostname=hostname)
        req.connection = DummyRequest(local_addr=(None, local_port))
        req.subprocess_env = {}
        if https:
            req.https = True
            req.subprocess_env['HTTPS'] = https

        return mkRequest(mpopenid.OpenIDAccessRequest, None, None, req)

    # Standard ports
    test_localhost80 = mkServerURLTest(
        ('localhost', 80), 'http://localhost/')
    test_localhost443_ssl = mkServerURLTest(
        ('localhost', 443, 'on'), 'https://localhost/')
    test_myopenid_com443_ssl = mkServerURLTest(
        ('myopenid.com', 443, 'on'), 'https://myopenid.com/')

    # Non-standard ports
    test_localhost8000 = mkServerURLTest(
        ('localhost', 8000), 'http://localhost:8000/')
    test_localhost8000_ssl = mkServerURLTest(
        ('localhost', 8000, 'on'), 'https://localhost:8000/')
    test_localhost80_ssl = mkServerURLTest(
        ('localhost', 80, 'on'), 'https://localhost:80/')

###### Redirection

class Redirect(Exception):
    pass

class DummyOpenIDRequest(mpopenid.OpenIDAccessRequest):
    def action_url(self, action):
        return 'action://' + action

    def redirect(self, url):
        raise Redirect(url)

def mkRedirectTest(message, target, options=None):
    if options is None:
        options = {}
    def test(self):
        req = mkRequest(DummyOpenIDRequest, {}, options, None)
        try:
            req.login_redirect(message, target)
        except Redirect, e:
            self.failUnlessEqual(e[0], 'action://login')
            self.failUnlessEqual(req.session['message'], message)
            if target is None:
                self.failUnlessRaises(
                    KeyError, req.session.__getitem__, 'target')
            else:
                self.failUnlessEqual(req.session['target'], target)
        else:
            self.fail('Expected redirect')
    return test

class RedirectTest(unittest.TestCase):
    test_loginRedirect = mkRedirectTest(None, None)
    test_loginRedirect_msg = mkRedirectTest('hi', None)
    test_loginRedirect_target = mkRedirectTest(None, 'target')
    test_loginRedirect_msg_target = mkRedirectTest('hi', 'target')

###### Creating a consumer

try:
    # python-openid 1.2, 2.x
    import openid.store.memstore
    teststore = openid.store.memstore.MemoryStore
    storetype = 'memory'
except ImportError:
    # python-openid 1.1
    import openid.store.dumbstore
    teststore = openid.store.dumbstore.DumbStore
    storetype = 'stateless'

class ConsumerTest(unittest.TestCase):
    def test_dumb(self):
        options = {}
        req = mkRequest(mpopenid.OpenIDLogin, None, options, None)
        self.failUnless(isinstance(req.consumer.consumer.store, teststore))

    def test_dumb_stateless(self):
        options = {'store-type':storetype}
        req = mkRequest(mpopenid.OpenIDLogin, None, options, None)
        self.failUnless(isinstance(req.consumer.consumer.store, teststore))

    def test_file(self):
        options = {'store-type':'file',
                   'store-directory':'/tmp/test-mpopenid-store/',
                   }
        req = mkRequest(mpopenid.OpenIDLogin, None, options, None)
        self.failUnless(isinstance(req.consumer.consumer.store,
                                   openid.store.filestore.FileOpenIDStore))

    def test_bad(self):
        options = {'store-type':'filex',
                   'store-directory':'/tmp/test-mpopenid-store/',
                   }
        errors = []
        def log_error(*args):
            errors.append(args)
        apache_request = DummyRequest(log_error=log_error)

        req = mkRequest(mpopenid.OpenIDLogin, None, options, apache_request)
        self.failUnlessRaises(SERVER_RETURN, req.get_consumer)
        self.failUnlessEqual(len(errors), 1)
        self.failUnless('filex' in errors[0][0])


###### Action Path
# options['action-path']
# apache_request.path_info
# apache_request.uri

def mkLoginActionPathTest(expected, uri, action_path=None, path_info=''):
    def test(self):
        options = {}
        if action_path:
            options['action-path'] = action_path

        apache_request = DummyRequest(path_info=path_info,
                               uri=uri)

        req = mkRequest(mpopenid.OpenIDLogin, None, options, apache_request)
        self.failUnlessEqual(req.action_path, expected)
    return test

class LoginActionPathTest(unittest.TestCase):
    test_empty = mkLoginActionPathTest('/', '/')

    test_basic = mkLoginActionPathTest('/openid/', '/openid')

    test_option = mkLoginActionPathTest('/foo/', '/openid/greeble', '/foo')

    test_pathinfo = mkLoginActionPathTest('/openid/', '/openid/', None, '/')

    test_pathinfo2 = mkLoginActionPathTest(
        '/openid/', '/openid/login', None, '/login')
    
    test_pathinfo_option = mkLoginActionPathTest(
        '/beans/', '/openid/login', '/beans', '/login')

    test_option_slash = mkLoginActionPathTest(
        '/beans/', '/openid/login', '/beans/', '/login')

###### .do_return

class DummyTable(dict):
    def add(self, key, value):
        self[key] = value

class DummyStorage(object):
    def __init__(self, req):
        if hasattr(req, 'query'):
            self.query = req.query
    def keys(self):
        return {}
    def getfirst(self, key, default=None):
        return self.query.get(key, default)

class MockConsumer(object):
    def __init__(self, status, **kwargs):
        self.kwargs = kwargs
        self.status = getattr(mpopenid.consumer, status)
    def complete(self, query, return_to):
        response = Dummy(status=self.status, **self.kwargs)
        if self.status == 'success':
            response.endpoint = Dummy(server_url='foo')
        return response
    def begin(self, op):
        pass

class MockConsumerOpenIDLogin(mpopenid.OpenIDLogin):
    def get_consumer(self):
        return self.consumer
    def set_last_user(self, usr):
        pass
    def get_allowed_ops(self):
        return {'a': 'frobble'}
    def fetch_op(self, endpoint):
        return 'an_allowed_op'

class DoReturnTest(unittest.TestCase):
    def setUp(self):
        from mod_python import util
        self.storage = util.FieldStorage
        util.FieldStorage = DummyStorage

    def tearDown(self):
        from mod_python import util
        util.FieldStorage = self.storage

    def request(self, session=None):
        if session is None:
            session = {}
        req = DummyRequest()
        return mkRequest(MockConsumerOpenIDLogin, session, {}, req)

    def test_do_return_failure(self):
        req = self.request()
        req.consumer = MockConsumer(status='FAILURE')
        self.assertRaises(SERVER_RETURN, req.do_return)
        location = req.apache_request.headers_out['location']
        self.assertEquals('http://localhost/login', location)

    def test_do_return_cancelled(self):
        req = self.request()
        req.consumer = MockConsumer(status='CANCEL')
        self.assertRaises(SERVER_RETURN, req.do_return)
        location = req.apache_request.headers_out['location']
        self.assertEquals('http://localhost/login', location)

    def test_do_return_setup_needed(self):
        req = self.request()
        req.consumer = MockConsumer(status='SETUP_NEEDED')
        self.assertRaises(SERVER_RETURN, req.do_return)
        location = req.apache_request.headers_out['location']
        self.assertEquals('http://localhost/login', location)

    def test_do_return_no_target(self):
        req = self.request()
        req.consumer = MockConsumer(status='SUCCESS', identity_url='a@b.c')
        self.assertRaises(SERVER_RETURN, req.do_return)
        location = req.apache_request.headers_out['location']
        self.assertEquals('http://localhost/login', location)

    def test_do_return_success(self):
        req = self.request(session={'target': 'foo'})
        req.consumer = MockConsumer(status='SUCCESS', identity_url='a@b.c')
        self.assertRaises(SERVER_RETURN, req.do_return)
        location = req.apache_request.headers_out['location']
        self.assertEquals('foo', location)

# .do_login
class DoLoginTest(unittest.TestCase):
    def setUp(self):
        from mod_python import util
        self.storage = util.FieldStorage
        util.FieldStorage = DummyStorage

    def tearDown(self):
        from mod_python import util
        util.FieldStorage = self.storage

    def request(self, session=None, query=None):
        if session is None:
            session = {}
        if query is None:
            query = {}
        self.req = DummyRequest(query=query)
        return mkRequest(MockConsumerOpenIDLogin, session, {}, self.req)

    def test_do_login_restricted(self):
        """ Check that the claimed openid_identifier is checked against the
            list of allowed openids """
        req = self.request(query={'openid_identifier': 'foo'})
        self.assertEquals(mpopenid.apache.OK, req.do_login())
        self.assertFalse('foo' in self.req.text)

    def test_do_login_success(self):
        """ Check that the claimed openid_identifier is included in the form
            if it's in the list of allowed openids """
        req = self.request(query={'openid_identifier': 'frobble'})
        self.assertEquals(mpopenid.apache.OK, req.do_login())
        self.assertTrue('frobble' in self.req.text)



###### Tests that need to be written
# mpopenid.OpenIDLogin
#  .openid
#  .do_logout
#  .fill_login_page
#
# mpopenid.OpenIDProtect
#  .get_authorized_urls
#  .get_action_path
#  .protect

if __name__ == '__main__':
    unittest.main()

